"""Constants for `TinCan Vocabulary` xAPI profile."""

# Context extension IRIs
CONTEXT_EXTENSION_PLANNED_DURATION = (  # pylint:disable=invalid-name
    "http://id.tincanapi.com/extension/planned-duration"
)
