"""Constants for `cmi5 Profile` xAPI profile."""

# Context extension IDs
CONTEXT_EXTENSION_SESSION_ID = (  # pylint:disable=invalid-name
    "https://w3id.org/xapi/cmi5/context/extensions/sessionid"
)
